from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains
from selenium.common.exceptions import NoSuchElementException, NoSuchFrameException, NoSuchAttributeException, \
    NoSuchWindowException, ElementNotSelectableException, ElementNotVisibleException, WebDriverException, \
    TimeoutException
from samuranium.logger import *
from samuranium.driver_factory import DriverFactory


class Core(object):
    def __init__(self):
        self.driver = None
        self.action_chains = ActionChains(self.driver)
        self.currentElement = None
        self.listOfElements = []
        self.desired_caps = {}
        self.driver_factory = DriverFactory()
        self.implicit_wait = 10

        super(Core, self)

    def set_driver(self, browser):
        """This method sets the driver the driver -- Should only be used by go_to_url(browser) method
        :param browser: name of the browser to be created
        """
        self.driver = self.driver_factory.set_driver(browser)

    def set_browser_type(self, browser):
        self.driver_factory.set_browser_type(browser)

    def go_to_url(self, url):
        if self.driver is None:
            self.driver = self.driver_factory.set_driver()
        try:

            log("Navigating to page: " + url)
            self.driver.get(url)
            return True
        except Exception as e:
            fail("Could not navigate to url: " + url)
            return False

    def get_current_page_url(self):
        return self.driver.current_url

    def take_screenshot(self, name):
        try:
            self.driver.get_screenshot_as_file(name + '.png')
            return True
        except IOError as e:
            error("Unable to save screenshot: " + e)
            return False

    # Methods to select objects
    def reset_current_element(self):
        self.currentElement = None
        return True

    def select_default_object(self):
        """Sets current element as the default element(Html/body) which contains all the elements in the HTML's body"""
        try:
            debug("Selecting root element")
            self.currentElement = self.driver.find_element_by_xpath("/html/body")
            return True
        except NoSuchElementException:
            error("Unable to select html/body element")
            return False

    def select_element_by_id(self, id):
        """Select an Element by ID
        :param id: element id value
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_id(id)
            else:
                self.currentElement = self.driver.find_element_by_id(id)
            debug("Element selected by id: " + id)
            return True
        except NoSuchElementException:
            error("Element with id: " + id + " not found")
            return False

    def select_element_by_class_name(self, class_name):
        """Select an Element by class name
        :param class_name: element class value
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_class_name(class_name)
            else:
                self.currentElement = self.driver.find_element_by_class_name(class_name)
            debug("Element selected by class: " + class_name)
            return True
        except NoSuchElementException:
            error("Element with class: " + class_name + " not found")
            return False

    def select_element_by_tag_name(self, tag_name):
        """Select an Element by tag name
        :param tag_name: element tag value
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_tag_name(tag_name)
            else:
                self.currentElement = self.driver.find_element_by_tag_name(tag_name)
            debug("Element selected by tag name: " + tag_name)
            return True
        except NoSuchElementException:
            error("Element with tag name: " + tag_name + " not found")
            return False

    def select_element_by_xpath(self, xpath):
        """Select an Element by Class Name
        :param xpath:
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_xpath(xpath)
            else:
                self.currentElement = self.driver.find_element_by_xpath(xpath)
            debug("Element selected by xpath: " + xpath)
            return True
        except NoSuchElementException:
            error("Element with xpath: " + xpath + " not found")
            return False

    def select_element_by_css_selector(self, css_selector):
        """Select an Element by Class Name
        :param css_selector:
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_css_selector(css_selector)
            else:
                self.currentElement = self.driver.find_element_by_css_selector(css_selector)
            debug("Element selected by css selector: " + css_selector)
            return True
        except NoSuchElementException:
            error("Element with css selector: " + css_selector + " not found")
            return False

    def select_element_by_link_text(self, link):
        """Select an Element by link text
        :param link:
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_link_text(link)
            else:
                self.currentElement = self.driver.find_element_by_link_text(link)
            debug("Element selected by link: " + link)
            return True
        except NoSuchElementException:
            error("Element with link: " + link + " not found")
            return False

    def select_element_by_partial_link_text(self, link):
        """Select an Element by partial link text
        :param link: element partial link text
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_partial_link_text(link)
            else:
                self.currentElement = self.driver.find_element_by_partial_link_text(link)
            debug("Element selected by partial link text: " + link)
            return True
        except NoSuchElementException:
            error("Element with partial link text: " + link + " not found")
            return False

    def select_element_by_name(self, name):
        """Select an Element by Class Name
        :param name: element name value
        """
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_name(name)
            else:
                self.currentElement = self.driver.find_element_by_name(name)
            debug("Element selected by name: " + name)
            return True
        except NoSuchElementException:
            error("Element with id: " + name + " not found")
            return False

    def select_parent_element(self):
        try:
            if self.currentElement is not None:
                self.currentElement.find_element_by_xpath("..")
            else:
                self.currentElement = self.currentElement.find_element_by_xpath("..")
            debug("Parent Element selected")
            return True
        except NoSuchElementException:
            error("Parent Element not found")
            return False

    # Implicit waits
    def wait_for_element_to_be_displayed_by_id(self, id):
        debug("Waiting for element by id to be displayed: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.ID, id)))
            debug("Element found by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    def wait_for_element_to_be_invisible_by_id(self, id):
        debug("Waiting for element by id to disappear: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.ID, id)))
            debug("Element by id disappeared: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(self.implicit_wait) + " seconds, the element by id is still visible: " + id)
            return False

    def wait_for_element_to_be_present_by_id(self, id):
        debug("Waiting for element by id to be present: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.ID, id)))
            debug("Element present by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    # Implicit waits
    def wait_for_element_to_be_displayed_by_id(self, id):
        debug("Waiting for element by id to be displayed: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.ID, id)))
            debug("Element found by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    def wait_for_element_to_be_invisible_by_id(self, id):
        debug("Waiting for element by id to disappear: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.ID, id)))
            debug("Element by id disappeared: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(self.implicit_wait) + " seconds, the element by id is still visible: " + id)
            return False

    def wait_for_element_to_be_present_by_id(self, id):
        debug("Waiting for element by id to be present: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.ID, id)))
            debug("Element present by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    # Implicit waits
    def wait_for_element_to_be_displayed_by_id(self, id):
        debug("Waiting for element by id to be displayed: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.ID, id)))
            debug("Element found by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    def wait_for_element_to_be_invisible_by_id(self, id):
        debug("Waiting for element by id to disappear: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.ID, id)))
            debug("Element by id disappeared: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(self.implicit_wait) + " seconds, the element by id is still visible: " + id)
            return False

    def wait_for_element_to_be_present_by_id(self, id):
        debug("Waiting for element by id to be present: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.ID, id)))
            debug("Element present by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    # Implicit waits - id
    def wait_for_element_to_be_displayed_by_id(self, id):
        debug("Waiting for element by id to be displayed: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.ID, id)))
            debug("Element found by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    def wait_for_element_to_be_invisible_by_id(self, id):
        debug("Waiting for element by id to disappear: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.ID, id)))
            debug("Element by id disappeared: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(self.implicit_wait) + " seconds, the element by id is still visible: " + id)
            return False

    def wait_for_element_to_be_present_by_id(self, id):
        debug("Waiting for element by id to be present: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.ID, id)))
            debug("Element present by id: " + id)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by id: " + id)
            return False

    # Implicit waits - class
    def wait_for_element_to_be_displayed_by_class_name(self, class_name):
        debug("Waiting for element by class name to be displayed: " + id)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.CLASS_NAME, class_name)))
            debug("Element found by class name: " + class_name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by class name: " + class_name)
            return False

    def wait_for_element_to_be_invisible_by_class_name(self, class_name):
        debug("Waiting for element by class name to disappear: " + class_name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.CLASS_NAME, class_name)))
            debug("Element by class name disappeared: " + class_name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(
                self.implicit_wait) + " seconds, the element by class name is still visible: " + class_name)
            return False

    def wait_for_element_to_be_present_by_class_name(self, class_name):
        debug("Waiting for element by class name to be present: " + class_name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.CLASS_NAME, class_name)))
            debug("Element present by class name: " + class_name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by class name: " + class_name)
            return False

    # Implicit waits - name
    def wait_for_element_to_be_displayed_by_name(self, name):
        debug("Waiting for element by name to be displayed: " + name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.NAME, name)))
            debug("Element found by name: " + name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by name: " + name)
            return False

    def wait_for_element_to_be_invisible_by_name(self, name):
        debug("Waiting for element by name to disappear: " + name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.NAME, name)))
            debug("Element by name disappeared: " + name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(self.implicit_wait) + " seconds, the element by id is still visible: " + id)
            return False

    def wait_for_element_to_be_present_by_name(self, name):
        debug("Waiting for element by name to be present: " + name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.NAME, name)))
            debug("Element present by name: " + name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by name: " + name)
            return False

    # Implicit waits - tag
    def wait_for_element_to_be_displayed_by_tag_name(self, tag_name):
        debug("Waiting for element by tag name to be displayed: " + tag_name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.TAG_NAME, tag_name)))
            debug("Element found by tag name: " + tag_name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by tag name: " + tag_name)
            return False

    def wait_for_element_to_be_invisible_by_tag_name(self, tag_name):
        debug("Waiting for element by tag name to disappear: " + tag_name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.TAG_NAME, tag_name)))
            debug("Element by tag name disappeared: " + tag_name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(
                    self.implicit_wait) + " seconds, the element by tag name is still visible: " + tag_name)
            return False

    def wait_for_element_to_be_present_by_tag_name(self, tag_name):
        debug("Waiting for element by tag name to be present: " + tag_name)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.TAG_NAME, tag_name)))
            debug("Element present by tag name: " + tag_name)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by tag name: " + tag_name)
            return False

    # Implicit waits - link text
    def wait_for_element_to_be_displayed_by_link_text(self, link_text):
        debug("Waiting for element by link text to be displayed: " + link_text)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.LINK_TEXT, link_text)))
            debug("Element found by link text: " + link_text)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by link text: " + link_text)
            return False

    def wait_for_element_to_be_invisible_by_link_text(self, link_text):
        debug("Waiting for element by link text to disappear: " + link_text)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.LINK_TEXT, link_text)))
            debug("Element by link text disappeared: " + link_text)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(
                    self.implicit_wait) + " seconds, the element by link text is still visible: " + link_text)
            return False

    def wait_for_element_to_be_present_by_link_text(self, link_text):
        debug("Waiting for element link text id to be present: " + link_text)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.LINK_TEXT, link_text)))
            debug("Element present by link text: " + link_text)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by link text: " + link_text)
            return False

    # Implicit waits - partial link text
    def wait_for_element_to_be_displayed_by_partial_link_text(self, link_text):
        debug("Waiting for element by partial link text to be displayed: " + link_text)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.PARTIAL_LINK_TEXT, link_text)))
            debug("Element found by partial link text: " + link_text)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by partial link text: " + link_text)
            return False

    def wait_for_element_to_be_invisible_by_partial_link_text(self, link_text):
        debug("Waiting for element by partial link text to disappear: " + link_text)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.PARTIAL_LINK_TEXT, link_text)))
            debug("Element by partial link text disappeared: " + link_text)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(
                    self.implicit_wait) + " seconds, the element by partial link text is still visible: " + link_text)
            return False

    def wait_for_element_to_be_present_by_partial_link_text(self, link_text):
        debug("Waiting for element partial link text id to be present: " + link_text)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.PARTIAL_LINK_TEXT, link_text)))
            debug("Element present by partial link text: " + link_text)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by partial link text: " + link_text)
            return False
        
    # Implicit waits - xpath
    def wait_for_element_to_be_displayed_by_xpath(self, xpath):
        debug("Waiting for element by xpath to be displayed: " + xpath)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.XPATH, xpath)))
            debug("Element found by xpath: " + xpath)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by xpath: " + xpath)
            return False

    def wait_for_element_to_be_invisible_by_xpath(self, xpath):
        debug("Waiting for element by xpath to disappear: " + xpath)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.XPATH, xpath)))
            debug("Element by xpath disappeared: " + xpath)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(self.implicit_wait) + " seconds, the element by id is still visible: " + id)
            return False

    def wait_for_element_to_be_present_by_xpath(self, xpath):
        debug("Waiting for element by xpath to be present: " + xpath)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.XPATH, xpath)))
            debug("Element present by xpath: " + xpath)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by xpath: " + xpath)
            return False
        
    # Implicit waits - css selector
    def wait_for_element_to_be_displayed_by_css_selector(self, css_selector):
        debug("Waiting for element by css_selector to be displayed: " + css_selector)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, css_selector)))
            debug("Element found by css_selector: " + css_selector)
            return True
        except (NoSuchElementException, TimeoutException):
            error("Unable to find element by css_selector: " + css_selector)
            return False

    def wait_for_element_to_be_invisible_by_css_selector(self, css_selector):
        debug("Waiting for element by css_selector to disappear: " + css_selector)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.invisibility_of_element_located(
                    (By.CSS_SELECTOR, css_selector)))
            debug("Element by css_selector disappeared: " + css_selector)
            return True
        except(NoSuchElementException, TimeoutException):
            error("After waiting " + str(self.implicit_wait) + " seconds, the element by id is still visible: " + id)
            return False

    def wait_for_element_to_be_present_by_css_selector(self, css_selector):
        debug("Waiting for element by css_selector to be present: " + css_selector)
        try:
            element = WebDriverWait(self.driver, self.implicit_wait).until(EC.presence_of_element_located(
                    (By.CSS_SELECTOR, css_selector)))
            debug("Element present by css_selector: " + css_selector)
            return True
        except(NoSuchElementException, TimeoutException):
            error("Unable to find element by css_selector: " + css_selector)
            return False

    # Methods to build a list of elements
    def build_element_list_by_id(self, id):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_id(id)
        else:
            self.listOfElements = self.driver.find_elements_by_id(id)
        if self.get_list_size() > 0:
            debug("List built by id: " + id)
            return True
        else:
            error("No elements found by id: " + id)
            return False

    def build_element_list_by_class_name(self, class_name):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_class_name(class_name)
        else:
            self.listOfElements = self.driver.find_elements_by_class_name(class_name)
        if self.get_list_size() > 0:
            debug("List built by id: " + class_name)
            return True
        else:
            error("No elements found by id: " + class_name)
            return False

    def build_element_list_by_tag_name(self, tag_name):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_tag_name(tag_name)
        else:
            self.listOfElements = self.driver.find_elements_by_tag_name(tag_name)
        if self.get_list_size() > 0:
            debug("List built by id: " + tag_name)
            return True
        else:
            error("No elements found by id: " + tag_name)
            return False

    def build_element_list_by_xpath(self, xpath):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_xpath(xpath)
        else:
            self.listOfElements = self.driver.find_elements_by_xpath(xpath)
        if self.get_list_size() > 0:
            debug("List built by id: " + xpath)
            return True
        else:
            error("No elements found by id: " + xpath)
            return False

    def build_element_list_by_css_selector(self, css_selector):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_css_selector(css_selector)
        else:
            self.listOfElements = self.driver.find_elements_by_css_selector(css_selector)
        if self.get_list_size() > 0:
            debug("List built by id: " + css_selector)
            return True
        else:
            error("No elements found by id: " + css_selector)
            return False

    def build_element_list_by_link_text(self, link):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_link_text(link)
        else:
            self.listOfElements = self.driver.find_elements_by_link_text(link)
        if self.get_list_size() > 0:
            debug("List built by id: " + link)
            return True
        else:
            error("No elements found by id: " + link)
            return False

    def build_element_list_by_partial_link_text(self, link):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_partial_link_text(link)
        else:
            self.listOfElements = self.driver.find_elements_by_partial_link_text(link)
        if self.get_list_size() > 0:
            debug("List built by id: " + link)
            return True
        else:
            error("No elements found by id: " + link)
            return False

    def build_element_list_by_name(self, name):
        if self.currentElement is not None:
            self.listOfElements = self.currentElement.find_elements_by_name(name)
        else:
            self.listOfElements = self.driver.find_elements_by_name(name)
        if self.get_list_size() > 0:
            debug("List built by id: " + name)
            return True
        else:
            error("No elements found by id: " + name)
            return False

    """Methods to select a specific element from listOfElements"""

    def get_list_size(self):
        return len(self.listOfElements)

    def select_element_from_list_by_index(self, index):
        if index <= self.get_list_size():
            self.currentElement = self.listOfElements[index]
            return True
        else:
            error("Element with the given index '" + str(index) + "' does not exist")
            return False

    def select_first_element_from_list(self):
        if self.get_list_size() > 0:
            self.currentElement = self.listOfElements[0]
            debug("Selected first element from list")
            return True
        else:
            error("List is emtpy")
            return False

    def select_last_element_from_list(self):
        if len(self.listOfElements) > 0:
            self.currentElement = self.listOfElements[self.get_list_size()]
            debug("Selected last element from list")
            return True
        else:
            error("List is emtpy")
            return False

    def select_element_from_list_by_attribute_and_value(self, attribute, value):
        found = False
        if self.get_list_size() > 0:
            for element in self.listOfElements:
                if value in element.get_attribute:
                    self.currentElement = element
                    found = True
                    debug("Element selected from list by attribute and value: " + attribute + " - " + value)
                    break
            if not found:
                error("Element not found by attribute and value: " + attribute + " - " + value)
        else:
            error("List is emtpy")
        return found

    # Methods to interact with the selected Object
    def type(self, value):
        debug("Typing text on current element: " + value)
        try:
            self.currentElement.clear()
            self.currentElement.send_keys(value)
            return True
        except Exception as e:
            error(e)
            return False

    def submit(self):
        debug("Pressing the Submit button on current element")
        try:
            self.currentElement.submit()
            return True
        except Exception as e:
            error(e)
            return False

    def click(self):
        debug("Clicked on current element")
        try:
            self.currentElement.click()
            return True
        except Exception as e:
            error(e)
            return False

    def hover(self):
        debug("Hovering over current element")
        try:
            hover = ActionChains(self.driver).move_to_element(self.currentElement)
            hover.perform()
            return True
        except Exception as e:
            error(e)
            return False

    def double_click(self):
        debug("Double clicked on current element")
        action_chains = ActionChains(self.driver)
        action_chains.double_click(self.currentElement).perform()

    def get_text(self):
        return self.currentElement.text

    def close(self):
        try:
            if self.driver is not None:
                log("Closing driver: " + self.driver.name)
                self.driver.quit()
                return True
        except WebDriverException as e:
            error(e)
            return False

    def get_tag_name(self):
        return self.currentElement.tag_name

    def get_id(self):
        return self.currentElement.id

    def get_location(self):
        return self.currentElement.location

    def get_location_once_scrolled_into_view(self):
        return self.currentElement.location_once_scrolled_into_view

    def get_size(self):
        return self.currentElement.size

    def is_displayed(self):
        return self.currentElement.is_displayed()

    def is_enabled(self):
        return self.currentElement.is_enabled()

    def is_selected(self):
        return self.currentElement.is_selected()

    def get_attribute(self, attribute):
        try:
            return self.currentElement.get_attribute(attribute)
        except NoSuchAttributeException:
            error("Attribute not found: " + attribute)
            return ""

    def get_css_property(self, prop):
        return self.currentElement.value_of_css_property(prop)
